function ownKeys(object, enumerableOnly) { var keys = Object.keys(object); if (Object.getOwnPropertySymbols) { var symbols = Object.getOwnPropertySymbols(object); if (enumerableOnly) symbols = symbols.filter(function (sym) { return Object.getOwnPropertyDescriptor(object, sym).enumerable; }); keys.push.apply(keys, symbols); } return keys; }

function _objectSpread(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; if (i % 2) { ownKeys(Object(source), true).forEach(function (key) { _defineProperty(target, key, source[key]); }); } else if (Object.getOwnPropertyDescriptors) { Object.defineProperties(target, Object.getOwnPropertyDescriptors(source)); } else { ownKeys(Object(source)).forEach(function (key) { Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key)); }); } } return target; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

/**
 * 参考 koa 的中间件和 redux 中间件
 */
import { isPromise, isFunction } from './checker';
import msg from './msg';
// Koa 洋葱模型
export var onionCompose = function onionCompose(middlewares) {
  return function (context, next) {
    var index = -1;
    return dispatch(0);

    function dispatch(i) {
      if (i <= index) {
        return Promise.reject(new Error('next() 被调用多次'));
      }

      index = i;
      var fn = middlewares[i];
      if (i === middlewares.length) fn = next;
      if (!fn) return Promise.resolve();
      var called = fn(context, function () {
        return dispatch(i + 1);
      });

      if (isPromise(called)) {
        return called;
      }

      throw msg.PROMISE_ERROR;
    }
  };
}; // 类似 Redux 的 pipe 模型
// pipeCompose([a, b, c])(x) ==> c(b(a(x)))

export var pipeCompose = function pipeCompose() {
  var pipes = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : [];

  if (pipes.length === 0) {
    return function (arg) {
      return arg;
    };
  }

  if (pipes.length === 1) {
    return isFunction(pipes[0]) ? pipes[0] : function (ctx) {
      return _objectSpread(_objectSpread({}, ctx), pipes[0]);
    };
  }

  return pipes.map(function (pipe) {
    return isFunction(pipe) ? pipe : function (ctx) {
      return _objectSpread(_objectSpread({}, ctx), pipe);
    };
  }).reduceRight(function (next, current) {
    return function (ctx) {
      return next(current(ctx));
    };
  });
};
export var processCompose = function processCompose() {
  var processors = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : [];
  return function () {
    var context = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
    processors.forEach(function (processor) {
      processor(context);
    });
    return context;
  };
}; // 合并 prop

export var reducedCompose = function reducedCompose(props) {
  return props.reduce(function (acc, prop) {
    return _objectSpread(_objectSpread({}, acc), prop);
  }, {});
};
export var mergedCompose = function mergedCompose(props) {
  var fn = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : function (key, res) {
    return res;
  };
  return props.reduce(function (acc, prop) {
    Object.keys(prop).forEach(function (key) {
      if (!acc[key]) {
        acc[key] = fn(key, prop[key], true);
      } else {
        // 出现冲突的话，也就是两个 key 一致的话，返回的是数组
        acc[key] = fn(key, Array.isArray(acc[key]) ? acc[key].concat(prop[key]) : [acc[key], prop[key]], false);
      }
    });
    return acc;
  }, {});
};
export var pipeMergedCompose = function pipeMergedCompose(props) {
  var fn = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : function (key, res) {
    return res;
  };

  var _props$reduce = props.reduce(function (acc, prop) {
    if (isFunction(prop)) {
      acc.pipes.push(prop);
    } else {
      acc.merged.push(prop);
    }

    return acc;
  }, {
    pipes: [],
    merged: []
  }),
      pipes = _props$reduce.pipes,
      _props$reduce$merged = _props$reduce.merged,
      merged = _props$reduce$merged === void 0 ? [] : _props$reduce$merged;

  return pipeCompose(pipes)(mergedCompose(merged, fn));
};
export var timelineCompose = function timelineCompose(timelines, native) {
  var extension = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
  return timelines.reduce(function (acc, timeline) {
    if (native[timeline] && extension[timeline]) {
      return acc.concat(native[timeline], extension[timeline]);
    }

    if (native[timeline]) {
      return acc.concat(native[timeline]);
    }

    if (extension[timeline]) {
      return acc.concat(extension[timeline]);
    }

    return acc;
  }, []);
};
export default {
  onionCompose: onionCompose,
  timelineCompose: timelineCompose,
  pipeCompose: pipeCompose,
  pipeMergedCompose: pipeMergedCompose,
  processCompose: processCompose,
  mergedCompose: mergedCompose,
  reducedCompose: reducedCompose
};